/*==============================================================================
#  $Id: convertguide.c,v 1.5 2004/05/28 23:00:08 moyekj Exp $
# Copyright (C) 2003 Kevin J. Moye <moyekj@yahoo.com>
#
# Author: Kevin J. Moye
# With contributions by Philip Van Baren
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# Run this program with --help option for more info.
# Program to generate recording schedule for ReplayTV 4000/4500/5000 DVR series.
# (Updated to work with 5.0 software)
# This program relies on xmltv TV_CHECK software for obtaining tv
# schedules.  Specifically, use "tv_grab_na" program to download programming
# schedule in xmltv format.
#
# NOTE: See the help (--help option) for info on channel mapping and many
#       other features.
==============================================================================*/

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>

#include "replaySchedule.h"
#include "guide.h"

extern int RTV_OS;
extern int replayversion;
extern int guideversion;
extern int DEBUG;

//**********************************************************************************************
// Retrieve the 4K Replay Guide and convert to replaySchedule structures
// and return list of structures with replay channel information.
//**********************************************************************************************

char *bitmapping(int flags, struct mapping *map) 
{
    static char buffer[1024];
    int i;
    buffer[0] = 0;
    for(i=0; map[i].name != NULL; i++) {
        if((flags & map[i].value) == map[i].value) {
             if(buffer[0] != 0) strcat(buffer,"; ");
             strcat(buffer, map[i].name);
        }
    }
    return buffer;
}

char *mapping(int flags, struct mapping *map)
{
    int i;
    for(i=0; map[i].name != NULL; i++) {
        if(flags == map[i].value)
             return map[i].name;
    }
    return "";
}

int convertReplayGuide(
      char *replayIP, struct Channel ***channels, int *numChan,
      int **CategoryIndex, char ***CategoryName, int *numCategories,
      int *allocChan, int *allocCategories, int *indexes, int *allocIndex,
      struct Show ***shows, int *numShows, int *allocShows
    )
{
    int i,base,count;
    unsigned long r;
    unsigned char * result;
    unsigned char * timestamp;
    unsigned long size;
    struct replay_guide_snapshot ss;
    struct Channel *chan;
    struct replay_channel *rc;
    struct replay_show *rs;
    struct replay_theme *rt;
    struct tm *tm;

    //Tell the user were up to something
    if(!quiet) fprintf(stderr,">> Extracting channel info from '%s'\n", replayIP);

    r = guide_client_get_snapshot(&result, &timestamp, &size, replayIP, "0");
    if (r) {
        printf("get_snapshot: error %lx\n", r);
        exit(-1);
    }
    parse_replay_guide_snapshot(&result, &ss);
    if(result) free(result);

    if(!quiet) {
       fprintf(stderr,">> Parsing Replay channel information for '%s' ...\n", replayIP);
    }

    replayversion = ss.guide_header.version;
    // guideversion is the major portion of the software version number
    // where guideversion==1 is OS==4, guideversion==2 is OS==5
    guideversion = replayversion & 0xFFFF;
    // Replayversion is the dotted portion of the software version number
    replayversion >>= 16;
    RTV_OS = 3 + guideversion;
    if(!quiet) fprintf(stderr,"   (%d.%d software version)\n", RTV_OS, replayversion);

    count = ss.group_data.num_categories;
    base = (*numCategories);

    (*numCategories) += count;
    (*indexes) += count;

    // Allocate / reallocate space for the new category information
    if((*allocIndex) < (*indexes)) {
        if(((*allocIndex) == 0) || ((*CategoryIndex) == NULL)) {
            (*CategoryIndex) = (int *)malloc(sizeof(int)*(*indexes));
        } else {
            (*CategoryIndex) = (int *)realloc((*CategoryIndex), sizeof(int)*(*indexes));
        }
        (*allocIndex) = (*indexes);
    }
    if((*allocCategories) < (*numCategories)) {
        if(((*allocCategories) == 0) || ((*CategoryName) == NULL)) {
            (*CategoryName) = (char **)malloc(sizeof(char *)*(*numCategories));
        } else {
            (*CategoryName) = (char **)realloc((*CategoryName), sizeof(char *)*(*numCategories));
        }
        (*allocCategories) = (*numCategories);
    }

    // Add the category info to the list
    // Note: this really is only meaningful if a single IP address is used with XML output
    // but for consistency we accept any case
    for(i=0; i<count; i++) {
        (*CategoryIndex)[base + i] = ss.group_data.index[i];
        (*CategoryName)[base + i] = strdup(ss.group_data.name_block + ss.group_data.name_offset[i]);
    }

    count = ss.guide_header.num_channels_1;
    base = (*numChan);

    (*numChan) += count;

    // Allocate / reallocate space for the new category information
    if((*allocChan) < (*numChan)) {
        if(((*allocChan) == 0) || ((*channels) == NULL)) {
            (*channels) = (struct Channel **)malloc(sizeof(struct Channel *)*(*numChan));
        } else {
            (*channels) = (struct Channel **)realloc((*channels), sizeof(struct Channel *)*(*numChan));
        }
        (*allocChan) = (*numChan);
    }

    for(i=0; i<count; i++) {
        chan = (struct Channel *)malloc(sizeof(struct Channel));
        rc = &(ss.replay_channels[i]);
        resetChannelStruct(chan);
        strcpy(chan->ip, replayIP);
        strcpy(chan->Type, mapping(rc->channel_type, replay_channel_type_mapping));
        chan->firstRun = ((rc->flags & 0x20)==0x20);
        strcpy(chan->CategoryName, bitmapping(rc->category, ss.group_data.category_bitmapping));
        if( strlen(chan->CategoryName) == 0 ) {
           strcpy(chan->CategoryName, "Uncategorized");
        }
        chan->channelID = rc->channel_id;
        strcpy(chan->Name, rc->label);
        fixWeirdChars(chan->Name);
        rs = &(rc->replay_show);
        strcpy(chan->ShowFlags, bitmapping(rs->flags, show_flags_bitmapping));
        if(rs->channel.isvalid) {
            chan->ChannelNum = rs->channel.headend_channel.tuning;
            strcpy(chan->ChannelName, rs->channel.headend_channel.name);
            chan->tmsid = rs->channel.headend_channel.tmsid;
        }
        if(rs->program.isvalid) {
            tm = localtime(&(rs->program.fixed_program_record.event_time));
            strftime(chan->Time, sizeof(chan->Time), "%Y%m%d%H%M%S", tm);
            if(rs->program.fixed_program_record.title_len > 1) {
                strcpy(chan->Title, rs->program.fixed_program_record.datablock
                                   + rs->program.fixed_program_record.title_offset);
                fixWeirdChars(chan->Title);
            }
            chan->Minutes = rs->program.fixed_program_record.minutes;
        }
        chan->Guaranteed = (rs->guaranteed != 0);
        strcpy(chan->Days, bitmapping(rc->days_of_week, day_of_week_bitmapping));
        {
            tm = localtime(&(rc->channel_id));
            strftime(chan->ChannelDate, sizeof(chan->ChannelDate), "%Y%m%d%H%M%S", tm);
        }
        chan->PriorityBoost = rc->priority;
        chan->PaddingBefore = rs->before_padding;
        chan->PaddingAfter = rs->after_padding;
        if((rc->channel_type == 2) || (rc->channel_type == 4)) { // Theme or Zone
            rt = &(rc->replay_theme);
            chan->Minutes = rt->minutes;
            strcpy(chan->Keywords, rt->keywords);
            fixWeirdChars(chan->Keywords);
            strcpy(chan->SearchKeys, bitmapping(rt->flags, theme_flags_bitmapping));
        }
        (*channels)[base + i] = chan;
    }

    // Parse the already recorded shows, if specified
    if(CONFIG.recordedHtml[0])
        count = ss.num_shows;
    else
        count = 0;

    // Allocate / reallocate space for the new category information
    if((*allocShows) < ((*numShows)+count)) {
        if(((*allocShows) == 0) || ((*shows) == NULL)) {
            (*shows) = (struct Show **)malloc(sizeof(struct Show *)*((*numShows)+count));
        } else {
            (*shows) = (struct Show **)realloc((*shows), sizeof(struct Show *)*((*numShows)+count));
        }
        (*allocShows) = (*numShows)+count;
    }

    for(i=0; i<count; i++) {
        struct Show *show = (struct Show *)malloc(sizeof(struct Show));
        // Reset show struct
        resetShowStruct(show);
        strcpy(show->ip, replayIP);

        rs = &(ss.replay_shows[i]);
        
        show->channelID = rs->channel_id;
        if(rs->channel.isvalid) {
            show->ChannelNum = rs->channel.headend_channel.tuning;
            strcpy(show->ChannelName, rs->channel.headend_channel.name);
            //show->tmsid = rs->channel.headend_channel.tmsid;
        }

        if(rs->program.isvalid) {
            struct fixed_program_record * p = &rs->program.fixed_program_record;

            strcpy(show->OtherInfo, bitmapping(p->flags, program_flags_mapping));
            fixWeirdChars(show->OtherInfo);

            show->ustart = p->event_time;
            show->Minutes = p->minutes;
            show->ustop = show->ustart + show->Minutes*60;
            if (p->title_len > 1) {
                strcpy(show->Title, p->datablock + p->title_offset);
                fixWeirdChars(show->Title);
            }
            if (p->episode_len > 1) {
                strcpy(show->Episode, p->datablock + p->episode_offset);
                fixWeirdChars(show->Episode);
            }
            if (p->description_len > 1) {
                strcpy(show->Description, p->datablock + p->description_offset);
                fixWeirdChars(show->Description);
            }
            if (p->actor_len > 1) {
                strcpy(show->Actors, p->datablock + p->actor_offset);
                fixWeirdChars(show->Actors);
            }
            if (p->guest_len > 1) {
                strcpy(show->Guests, p->datablock + p->guest_offset);
                fixWeirdChars(show->Guests);
            }
            //if (p->suzuki_len > 1) {
            //    strcpy(show->Suzuki, p->datablock + p->suzuki_offset);
            //    fixWeirdChars(show->Suzuki);
            //}
            if (p->producer_len > 1) {
                strcpy(show->Producers, p->datablock + p->producer_offset);
                fixWeirdChars(show->Producers);
            }
            if (p->director_len > 1) {
                strcpy(show->Directors, p->datablock + p->director_offset);
                fixWeirdChars(show->Directors);
            }
            if(p->title_len > 1) {
                strcpy(show->Title, p->datablock + p->title_offset);
                fixWeirdChars(show->Title);
            }
            (*shows)[(*numShows)++] = show;
            if(DEBUG) {
                printf("**** Recorded show at index %d of %d\n",i+1,count);
                printRecShowInfo(show);
            }
        } else {
            if(DEBUG) {
                printf("**** Blank or invalid show found at index %d of %d\n",i+1,count);
            }
            free(show);
        }
    }
    free_replay_guide_snapshot(&ss);
    return 1;
}
