/*==============================================================================
# Copyright (C) 2003 Kevin J. Moye <moyekj@yahoo.com>
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# $Id: parseDataDirect.c,v 1.7 2005/05/20 18:52:48 moyekj Exp $
# $Log: parseDataDirect.c,v $
# Revision 1.7  2005/05/20 18:52:48  moyekj
# Missing station id is now warning instead of error
#
# Revision 1.6  2004/12/10 01:16:09  moyekj
# Changed keyword determining dd format
#
# Revision 1.5  2004/10/22 00:47:13  moyekj
# Added hdtv flag check
#
# Revision 1.4  2004/06/09 17:17:20  moyekj
# Added checking of dd file
#
# Revision 1.3  2004/03/31 05:51:21  moyekj
# dst fix to datadirect parsing
#
# Revision 1.2  2004/03/09 20:58:45  moyekj
# Proper start/stop formatting for xml output
#
# Revision 1.1  2004/03/09 01:15:57  moyekj
# Initial checkin.
#
==============================================================================*/

#define DEBUG_DD 0

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>

#include "replaySchedule.h"
#include "hash.h"

static char _time_[50];

// Station ID => name/num mapping structure
struct CHMAP_DD {
   char name[512];
   int  num;
};
htab *chHash;

// Add new channel id entry to global chHash hash
void chHashNew(char *name) {
 ub1  *key;
 ub4  keyl;
 struct CHMAP_DD *data;

   keyl = strlen(name)+1;
   hadd(chHash, name, keyl, (void *)0);
   key = (ub1 *)malloc(keyl);
   memcpy(key, name, keyl);
   hkey(chHash)=key;

   data = (struct CHMAP_DD *)malloc(sizeof(struct CHMAP_DD));
   hstuff(chHash)=data;
}

// Point channel hash location to given id entry
void chHashFind(char *id) {
   if( ! hfind(chHash, id, strlen(id)+1) ) {
      fprintf(stderr, "*WARNING* Cannot find station id: %s\n", id);
      id = (char *)NULL;
      //exit(1);
   }
}

// Find CHMAP_DD structure based on channel ID
struct CHMAP_DD *findChanDD(char *id) {
   if( hfind(chHash, id, strlen(id)+1) ) {
      return(hstuff(chHash));
   }
   fprintf(stderr, "*ERROR* Cannot find station id: %s\n", id);
   exit(1);
}

// Program ID structure
struct PROGMAP_DD {
   char         title[MAXLINE];
   char         subtitle[MAXLINE];
   char         desc[MAXLINE];
};
htab *progHash;

// Initialize program id structure
void resetProg(struct PROGMAP_DD *p)
{
   strcpy(p->title, "");
   strcpy(p->subtitle, "");
   strcpy(p->desc, "");
}

// Add new channel id entry to global progHash hash
void progHashNew(char *name) {
 ub1  *key;
 ub4  keyl;
 struct PROGMAP_DD *data;

   keyl = strlen(name)+1;
   hadd(progHash, name, keyl, (void *)0);
   key = (ub1 *)malloc(keyl);
   memcpy(key, name, keyl);
   hkey(progHash)=key;

   data = (struct PROGMAP_DD *)malloc(sizeof(struct PROGMAP_DD));
   resetProg(data);
   hstuff(progHash)=data;
}

// Point program hash location to given id entry
void progHashFind(char *id) {
   if( ! hfind(progHash, id, strlen(id)+1) ) {
      fprintf(stderr, "*ERROR* Cannot find program id: %s\n", id);
      exit(1);
   }
}

// Find PROGMAP_DD structure based on program ID
struct PROGMAP_DD *findProgDD(char *id) {
   if( hfind(progHash, id, strlen(id)+1) ) {
      return(hstuff(progHash));
   }
   fprintf(stderr, "*ERROR* Cannot find program id: %s\n", id);
   exit(1);
}

void printDDProgram(struct prog *program)
{
   printf("**"); printProgramStruct(program);
}

// Fix xml schedule weird "'" representation (1st occurrence)
void fixApostrophe( char *string ) {
int i, j;
char *ptr, _string_[MAXLINE];
   // Replace "&apos;" with "'" in a string
   if( (ptr=strstr(string,"&apos;")) ) {
      i=0; j=0;
      // Copy into _string_ until start of ptr
      for( i=0; i<strlen(string)-strlen(ptr); ++i ) {
         _string_[j++] = string[i];
      }

      // Replace weird chars with just "'"
      _string_[j++] = '\'';

      // Copy remaining into _string_ after weird chars
      for( i=6; i<strlen(ptr); ++i ) {
         _string_[j++] = ptr[i];
      }
      _string_[j] = '\0';

      strcpy(string, _string_);
   }
}

// Fix xml schedule weird "'" representation (all occurrences)
char *fixApostrophes( char *string ) {
        static char s[1000];
   strcpy(s, string);
   while( strstr(s,"&apos;") ) {
      fixApostrophe(s);
   }
   return(s);
}

void ddScanErr(char *line, char *string)
{
   fprintf(stderr, "*ERROR* DataDirect scan fatal error:\n");
   fprintf(stderr, "   scanning for: '%s'\n",string);
   fprintf(stderr, "   line: '%s'\n",line);
   exit(1);
}

// convert given DataDirect start time to format: yyyyMMddhhmmss
// DataDirect start time format is: yyyy-MM-ddThh:mm:ssZ
char *ddStart(char *start)
{
 int i=0, dd=0, j, k;
 time_t gmtoffset, ustart, t1, t2;
 struct tm *T1, *T2;

   // Year
   for(j=0; j<4; ++j)
      _time_[i++] = start[dd++];

   // Rest
   for(k=0; k<5; ++k) {
      dd++;
      for(j=0; j<2; ++j)
         _time_[i++] = start[dd++];
   }
   _time_[i] = '\0';

   // Compute GMT start time in seconds
   ustart = guideTimeToUTC(_time_);

   // Compute GMT offset in seconds
   T1 = localtime(&ustart); T1->tm_isdst=-1; t1=mktime(T1);
   T2 = gmtime(&ustart); T2->tm_isdst=-1; t2=mktime(T2);
   gmtoffset = t1 - t2;

   // Compute local start time in seconds
   ustart += gmtoffset;

   // Get tm struct based on ustart
   T1 = localtime(&ustart); T1->tm_isdst=-1;

   // Want start format: yyyyMMddhhmmss
   sprintf(_time_, "%04d%02d%02d%02d%02d%02d",
      T1->tm_year+1900, T1->tm_mon+1, T1->tm_mday,
      T1->tm_hour, T1->tm_min, T1->tm_sec
   );

   return(_time_);
}

// convert given DataDirect start time + duration to stop time
// start format is: yyyyMMddhhmmss
// duration format is: PThhHmmM
// desired stop format is same as start format
char *ddStop(char *start, char *duration)
{
 char string[56];
 int i, dd=0;
 int hours, mins;
 time_t ustart, ustop;
 struct tm *T;

   // Collect hours from duration
   i=0;
   dd += 2;
   string[i++] = duration[dd++];
   string[i++] = duration[dd++];
   string[i] = '\0';
   hours = atoi(string);

   // Collect mins from duration
   i=0;
   dd++;
   string[i++] = duration[dd++];
   string[i++] = duration[dd++];
   string[i] = '\0';
   mins = atoi(string);

   // Compute stop time in UTC seconds
   ustart = guideTimeToUTC(start);
   ustop = ustart + hours*60*60 + mins*60;

   // Grab Time structure corresponding to stop UTC
   T = localtime(&ustop); T->tm_isdst=-1;

   // Want stop format: yyyyMMddhhmmss
   sprintf(_time_, "%04d%02d%02d%02d%02d%02d",
      T->tm_year+1900, T->tm_mon+1, T->tm_mday,
      T->tm_hour, T->tm_min, T->tm_sec
   );

   return(_time_);
}

// Scan the dd xml file & store all program info in progHash hash
void ddProgInfo(char ddScheduleFile[])
{
   FILE *ifp;
   char	line[MAXLINE], s[MAXLINE], token[MAXLINE];
   char *s1;
   int  i;
   struct PROGMAP_DD *pinfo;

   ifp = fopen(ddScheduleFile, "r");
   if( ! ifp ) {
      fprintf(stderr, "*ERROR* Failed to open file: '%s'\n", ddScheduleFile);
      exit(1);
   }

   progHash = hcreate(8);
   pinfo = (struct PROGMAP_DD *)malloc(sizeof(struct PROGMAP_DD));

   while( fgets(line, MAXLINE, ifp) ) {
      i = strlen(line)-1;
      if( line[i] == '\n' ) {
         line[i] = '\0';
      }

      // Parse the xml file, 1 line at a time
      // New program definition
      strcpy(token, "<program id='");
      if( hasStr(line, token, 1) ) {
         // Get program ID & add new program info hash entry
         s1 = strstr(line, token); if( !s1 ) ddScanErr(line, token);
         sscanf(s1+strlen(token), "%[^']", s);
         progHashNew(s); pinfo = hstuff(progHash);
      }

      // Title
      else if( (s1 = strstr(line, "<title>")) ) {
         sscanf(s1+strlen("<title>"), "%[^<]", s);
         strcpy(pinfo->title, fixApostrophes(fixAmpersans(s)));
      }

      // Sub-title
      else if( (s1 = strstr(line, "<subtitle>")) ) {
         sscanf(s1+strlen("<subtitle>"), "%[^<]", s);
         strcpy(pinfo->subtitle, fixApostrophes(fixAmpersans(s)));
      }

      // Description
      else if( (s1 = strstr(line, "<description>")) ) {
         sscanf(s1+strlen("<description>"), "%[^<]", s);
         strcpy(pinfo->desc, fixApostrophes(fixAmpersans(s)));
      }
   }
   fclose(ifp);
}

// Extract recording schedule from DataDirect XML
int extractRecordingScheduleDD(
   char ddScheduleFile[], struct Channel **channels, int numChan,
   struct Channel ***progs, int *pnum, int *numProg,
   struct stopTime *stopTimes )
{

   FILE *ifp;
   char	line[MAXLINE], s[MAXLINE], token[MAXLINE];
   char *s1, **unique=NULL;
   int  i, p=0, count=0;
   int  uniqueNum=0;
   int  isDD=0;

   struct Channel *match;
   struct prog program, progList[3];
   struct CHMAP_DD *chmp;
   struct PROGMAP_DD *pinfo;

   chHash = hcreate(8);
   chmp = (struct CHMAP_DD *)malloc(sizeof(struct CHMAP_DD));
   pinfo = (struct PROGMAP_DD *)malloc(sizeof(struct PROGMAP_DD));

   // Pass 1 - Collect program id -> title/subtitle/desc info
   ddProgInfo(ddScheduleFile);

   // Pass 2 - Build recording schedule
   resetProgStruct(&program);
   for(i=0;i<3;i++)
	   resetProgStruct(&progList[i]);

   if(!quiet) fprintf(stderr,">> Parsing '%s' (using DataDirect format) to determine schedule ...\n", ddScheduleFile);

   ifp = fopen(ddScheduleFile, "r");
   if( ! ifp ) {
      fprintf(stderr, "*ERROR* Failed to open file: '%s'\n", ddScheduleFile);
      return(0);
   }

   while( fgets(line, MAXLINE, ifp) ) {
      i = strlen(line)-1;
      if( line[i] == '\n' ) {
         line[i] = '\0';
      }

      // Parse the xml file, 1 line at a time
      if( hasStr(line,"xtvdResponse",1) ) {
         isDD = 1;
      }

      // <station id => add to channel map
      strcpy(token, "<station id='");
      if( hasStr(line, token, 1) ) {
         s1 = strstr(line, token); if( !s1 ) ddScanErr(line, token);
         sscanf(s1+strlen(token), "%[^']", s);
         chHashNew(s); chmp = hstuff(chHash);
      }

      // <callSign> => channel name
      strcpy(token, "<callSign>");
      if( hasStr(line, token, 1) ) {
         s1 = strstr(line, token); if( !s1 ) ddScanErr(line, token);
         sscanf(s1+strlen(token), "%[^<]", s);
         strcpy(chmp->name, s);
      }

      // <map station=' => channel num
      strcpy(token, "<map station='");
      if( hasStr(line, token, 1) ) {
         s1 = strstr(line, token); if( !s1 ) ddScanErr(line, token);
         sscanf(s1+strlen(token), "%[^']", s);
         chHashFind(s); chmp = hstuff(chHash);
         if(s) {
            strcpy(token, " channel='");
            s1 = strstr(line, token); if( !s1 ) ddScanErr(line, token);
            sscanf(s1+strlen(token), "%[^']", s);
            chmp->num = atoi(s);
         }
      }

      // <schedule program => new program schedule
      strcpy(token, "<schedule program='");
      if( hasStr(line, token, 1) ) {
         // Reset program structure
         resetProgStruct(&program);

         // program id
         s1 = strstr(line, token); if( !s1 ) ddScanErr(line, token);
         sscanf(s1+strlen(token), "%[^']", s);
         progHashFind(s); pinfo = hstuff(progHash);

         // station => store channel name & num
         strcpy(token, " station='");
         s1 = strstr(line, token); if( !s1 ) ddScanErr(line, token);
         sscanf(s1+strlen(token), "%[^']", s);
         chmp = findChanDD(s);
         strcpy(program.channel, channelMap(chmp->name));
         program.channelNum = chmp->num;

         // time => start time
         strcpy(token, " time='");
         s1 = strstr(line, token); if( !s1 ) ddScanErr(line, token);
         sscanf(s1+strlen(token), "%[^']", s);
         strcpy(program.start, ddStart(s));
         program.ustart = guideTimeToUTC(program.start);

         // duration => stop time
         strcpy(token, " duration='");
         s1 = strstr(line, token); if( !s1 ) ddScanErr(line, token);
         sscanf(s1+strlen(token), "%[^']", s);
         strcpy(program.stop, ddStop(program.start, s));
         program.ustop = guideTimeToUTC(program.stop);

         // repeat => flag as repeat
         strcpy(token, " repeat='true'");
         s1 = strstr(line, token);
         if( s1 ) program.repeat = 1;

         // hdtv flag
         strcpy(token, " hdtv='true'");
         s1 = strstr(line, token);
         if( s1 ) program.hdtv = 1;

         // Copy remaining info
         strcpy(program.title, pinfo->title);
         strcpy(program.subtitle, pinfo->subtitle);
         strcpy(program.desc, pinfo->desc);

         // Remember the earliest and latest start times
         // These are used for displaying single recordings within the
         // schedule times which don't have a matching program
         if( (program.ustart < minstarttime) || (minstarttime == 0) )
            minstarttime = program.ustart;
         if( (program.ustart > maxstarttime) || (maxstarttime == 0) )
            maxstarttime = program.ustart;

         if(DEBUG_DD) printDDProgram(&program);
         count++;
         if( count == 1 ) {
            progList[1] = program;
         }
         else if( count == 2 ) {
            progList[0] = progList[1];
            progList[1] = program;
         }
         else {
            progList[0] = progList[1];
            progList[1] = progList[2];
            progList[2] = program;
         }
         p++;

         // Process when progList has at least 2 entries
         if ( count > 1 ) {
            match = compareProgramToChannels(progList, channels, numChan);

            if(match) {
               if( match->title[0] && ensureUniqueProgram(&unique, &uniqueNum, match) ) {
                  // Allocate more memory for progs if needed
                  if ( *numProg >= *pnum ) {
                     *pnum += 100;
                     (*progs) = (struct Channel **)realloc(
                        (*progs), sizeof(struct Channel *)*(*pnum)
                     );
                  }
                  (*progs)[(*numProg)++] = match;
               } else {
                  free(match);
               }
            }
         }
      }
   }

   fclose(ifp);

   // 1 more iteration on progList needed (to get last program)
   match = compareProgramToChannels(progList, channels, numChan);

   if ( match && strlen(match->title) > 0 ) {
      if ( ensureUniqueProgram(&unique, &uniqueNum, match) ) {
         if ( *numProg >= *pnum ) {
            (*pnum) += 100;
            (*progs) = (struct Channel **)realloc(
               (*progs), sizeof(struct Channel *)*(*pnum)
            );
         }
         (*progs)[(*numProg)++] = match;
      }
   }

   if( ! isDD ) {
      fprintf(stderr, "*ERROR* %s not in DataDirect format... aborting\n",
         ddScheduleFile
      );
      exit(1);
   }

   // Free up the allocated memory
   if(unique) {
      for(i=(uniqueNum-1); i>=0; i--) {
         if(unique[i]) free(unique[i]);
      }
      free(unique);
   }
   free(chmp);
   free(pinfo);
   hdel(chHash);
   hdel(progHash);

   return(1);
}
